import re
import struct


class RGB8:
    @staticmethod
    def parse(color):
        """Parse a color as an RGB8 value."""
        short = re.compile(r"\A[#]*[a-fA-F0-9]{3}\Z")
        full = re.compile(r"\A[#]*[a-fA-F0-9]{6}\Z")

        if short.fullmatch(color):
            color = "".join([c * 2 for c in color.replace("#", "")])

        if full.fullmatch(color):
            return struct.unpack("BBB", bytes.fromhex(color.replace("#", "")))

        raise ValueError(f"{color} does not represent a valid RGB8 color value.")



import base64
import enum
import io
import logging
import string

import numpy as np
import PIL.Image

from .color import RGB8

logger = logging.getLogger(__name__)


class Resolutions(enum.Enum):
    """Enum that defines some common image resolutions

    Members of this enum are tuples containing the width and height which can be
    accessed by name::

       >>> import arlunio as st

       >>> hd = st.Resolutions.HD
       >>> hd.width
       1280

       >>> hd.height
       720

    Resolutions can also unpacked::

       >>> width, height = hd
       >>> width
       1280

       >>> height
       720

    Members are also available via the :code:`R` shortcut::

       >>> st.R.HD
       <Resolutions.HD: (1280, 720)>

    """

    HD = (1280, 720)
    """1280 x 720"""

    FHD = (1920, 1080)
    """1920 x 1080"""

    QHD = (2560, 1440)
    """2560 x 1440"""

    def __iter__(self):
        value = self.value
        return iter([value[0], value[1]])

    @property
    def width(self):
        return self.value[0]

    @property
    def height(self):
        return self.value[1]


class Image:
    """An image is a container for raw pixel data."""

    def __init__(self, pixels, mask=None):
        self.pixels = pixels
        self._mask = mask

    def __repr__(self):
        y, x, _ = self.pixels.shape
        return f"Image<{x} x {y}>"

    def _repr_html_(self):

        data = self.encode().decode("utf-8")
        html = """\
            <style>
              .arlunio-image {
                  width: 50%;
                  margin: auto;
                  image-rendering: crisp-edges;
                  image-rendering: pixelated;
                  border: solid 1px #ddd;
              }
            </style>
            <img class="arlunio-image" src="data:image/png;base64,$data"></img>
        """
        template = string.Template(html)

        return template.safe_substitute({"data": data})

    def __getitem__(self, key):
        return Image(self.pixels[key])

    def __setitem__(self, key, value):
        self.pixels[key] = value

    @property
    def mask(self):

        if self._mask is None:
            return tuple([slice(None, None, None) for _ in range(3)])

        return self._mask

    @mask.setter
    def mask(self, value):
        self._mask = value

    @classmethod
    def new(cls, width: int, height: int, background: str = None, colorspace=None):
        """Create a new Image with the given width and height.

        This creates an "empty" image of a given width and height with a solid
        background color. This color can be set using the :code:`background` color
        argument, or if :code:`None` then the background will default to white.

        The :code:`background` argument should be in the form of a string
        representing the color as an RGB hex code (like those used in web design
        e.g. :code:`#ffbb00`)

        The :code:`colorspace` parameter can be used to change the colorspace used when
        drawing the image. By default this is the :code:`RGB8` colorspace.

        :param width: The width of the image in pixels
        :param height: The height of the image in pixels
        :param background: The background color to use.
        :param colorspace: The colorspace to use.

        """

        if background is None:
            background = "ffffff"

        if colorspace is None:
            colorspace = RGB8

        bg_color = colorspace.parse(background)

        pixels = np.full((height, width, 3), bg_color, dtype=np.uint8)
        return cls(pixels)

    def _as_pillow_image(self):
        height, width, _ = self.pixels.shape

        return PIL.Image.frombuffer(
            "RGB", (width, height), self.pixels, "raw", "RGB", 0, 1
        )

    def save(self, filename: str) -> None:
        """Save an image in PNG format.

        :param filename: The filepath to save the image to.
        """

        image = self._as_pillow_image()

        with open(filename, "wb") as f:
            image.save(f)

    def encode(self) -> bytes:
        """Return the image encoded as a base64 string."""
        logger.debug("Encoding image as base64")
        image = self._as_pillow_image()

        with io.BytesIO() as byte_stream:
            image.save(byte_stream, "PNG")
            image_bytes = byte_stream.getvalue()

            return base64.b64encode(image_bytes)



import logging

import arlunio as ar
import numpy as np

logger = logging.getLogger(__name__)
pattern = ar.ShapeCollection()


@pattern.shape
def Grid(x, *, nx=4, ny=4, shape=None):
    """Repeatedly draw the given shape in a grid.

    :param nx: The number of times to repeat the shape in the x-direction
    :param ny: The number of times to repeat the shape in the y-direction
    :param shape: The shape instance to draw.
    """

    # Use a parameter to get the dimensions of the image we are drawing.
    # Remember! Parameters return an array of (height, width)
    height, width = x.shape
    bg = np.full((height, width), False)

    logger.debug(f"Grid size: {ny} x {nx}")
    logger.debug(f"Background size: {height} x {width}")

    # Draw the shape at a size determined by the size of the grid
    s_height, s_width = height // ny, width // nx
    mask = shape.mask(s_width, s_height)

    logger.debug(f"Shape size: {s_height} x {s_width}")

    # Let numpy handle the repeating of the shape across the image.
    pattern = np.tile(mask, (ny, nx))

    # Apply the pattern to the background, depending on the grid size and
    # image dimensions align, the generated grid may not perfectly fill the
    # image.
    p_height, p_width = pattern.shape
    logger.debug(f"Pattern size: {p_height} x {p_width}")

    bg[:p_height, :p_width] = pattern

    return bg



import textwrap

from arlunio.doc.directives import format_error, render_image
from docutils import nodes
from docutils.statemachine import StringList


class TestRenderImage:
    """Tests for the :code:`render_image` function."""

    def test_handles_syntax_error(self):
        """Ensure that if there is an error when compiling the code we can
        handle it."""

        src = "x 1 + 2"
        doctree = render_image(src)

        assert len(doctree) == 1
        assert isinstance(doctree[0], nodes.error), "Expected error node."

        error = doctree[0]
        assert len(error.children) == 2, "Expected two content nodes."

        text, trace = error.children

        assert isinstance(text, nodes.Text), "Expected text node."
        assert "Unable to render image" in text.astext()

        assert isinstance(trace, nodes.literal_block)
        assert "Traceback" in trace.astext()
        assert "SyntaxError" in trace.astext()

    def test_handles_runtime_error(self):
        """Ensure that if there is an error when executing the code we can handle it."""

        src = "x = 1 / 0"
        doctree = render_image(src)

        assert len(doctree) == 1
        assert isinstance(doctree[0], nodes.error), "Expected error node."

        error = doctree[0]
        assert len(error.children) == 2, "Expected two content nodes."

        text, trace = error.children

        assert isinstance(text, nodes.Text), "Expected text node."
        assert "Unable to render image" in text.astext()

        assert isinstance(trace, nodes.literal_block)
        assert "Traceback" in trace.astext()
        assert "ZeroDivisionError" in trace.astext()

    def test_image_provided(self):
        """Ensure that if the code provides an image we use that."""

        src = """\
        import arlunio as ar

        circle = ar.S.Circle()
        disk = circle(4,4)
        """
        doctree = render_image(textwrap.dedent(src))

        assert len(doctree) == 1
        assert isinstance(doctree[0], nodes.raw), "Expected raw node"


class TestFormatError:
    """Tests for the :code:`format_error` function."""

    def test_formatter(self):

        err = "\n".join(
            [
                "Traceback (most recent call last):",
                '   File "<stdin>", line 1, in <module>',
                "ZeroDivisionError: division by zero",
            ]
        )
        content = format_error("Unable to perform :code:`2 / 0`", err)

        assert isinstance(content, StringList)

        actual = list(content)
        expected = [
            ".. error::",
            "",
            "   Unable to perform :code:`2 / 0`::",
            "",
            "      Traceback (most recent call last):",
            '         File "<stdin>", line 1, in <module>',
            "      ZeroDivisionError: division by zero",
            "",
        ]

        assert expected == actual



import logging

import pkg_resources

logger = logging.getLogger(__name__)
_LOAD_CACHE = {}


class Collection:
    """A collection of items, populate from an entry point."""

    def __init__(self, name, items):
        self.name = name
        self.items = items

    def __repr__(self):
        return f"Collection<{self.name}, {len(self.items)} items>"

    def _get(self, name):
        """Get an item, return None if none exists."""
        if name not in self.items:
            logger.debug(f"Item {name} does not exist in collection {self.name}")
            return None

        item = self.items[name]

        if isinstance(item, pkg_resources.EntryPoint):
            logger.debug(f"Item {name} is an entry point reference, loading...")

            real_item = item.load()
            self.items[name] = real_item

            return real_item

        return item

    def __getitem__(self, name):
        item = self._get(name)

        if item is None:
            raise KeyError(name)

        return item

    def __getattr__(self, name):
        item = self._get(name)

        if item is None:
            raise AttributeError(name)

        return item

    def list(self):
        """List all of the available items in the collection."""
        return sorted(list(self.items.keys()))


def _load_collection(name, entry_point, docstring=None):
    """This method builds an enum containing all the items defined in the named
    entry point.

    :param name: The name to give the constructed enum
    :param entry_point: The name of the entry point to load
    :param docstring: If given the constructed enum will be given the docstring
    """

    if name in _LOAD_CACHE:
        return _LOAD_CACHE[name]

    logger.debug(f"Loading {name}")
    items = {}

    for item in pkg_resources.iter_entry_points(entry_point):
        logger.debug(f"--> {item.module_name}.{item.name}")
        items[item.name] = item

    collection = Collection(name, items)

    if docstring is not None:
        collection.__doc__ = docstring

    _LOAD_CACHE[name] = collection
    return collection


def load_parameters():
    """Load all of the available parameters."""

    docstring = """\
    All of the available parameters.
    """

    return _load_collection("Parameter", "arlunio.parameters", docstring)
